from dataclasses import dataclass
import json
from pathlib import Path
from typing import Any, Iterator, Self

from kisiac.common import UserError, check_type, run_cmd


@dataclass(frozen=True)
class Encryption:
    name: str | None
    device: Path
    hash: str
    cipher: str
    key_size: int | None

    def open(self, host: str, password: str) -> None:
        assert self.name is not None
        run_cmd(
            ["cryptsetup", "open", "--key-file", "-", str(self.device), self.name],
            sudo=True,
            input=password,
        )

    def close(self, host: str) -> None:
        assert self.name is not None
        run_cmd(["cryptsetup", "close", self.name], sudo=True)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Encryption):
            return False
        # TODO remove this method once we know how to obtain
        # key_size from the system
        return (
            self.name == other.name
            and self.device == other.device
            and self.hash == other.hash
            and self.cipher == other.cipher
        )


@dataclass
class EncryptionSetup:
    encryptions: set[Encryption]

    @classmethod
    def from_config(cls, config: list[dict[str, Any]]) -> Self:
        check_type("encryption key", config, list)
        encryptions = set()
        for i, settings in enumerate(config):
            check_type(f"encryption item {i}", settings, dict)
            mapping = settings.get("mapping", {})
            check_type(f"mapping of encryption item {i}", mapping, dict)
            for name, device in mapping.items():
                check_type(f"device of encryption item {i}", device, str)
                check_type(f"name of encryption item {i}", name, str)
                try:
                    encryptions.add(
                        Encryption(
                            name=name,
                            device=Path(device),
                            hash=settings["hash"],
                            cipher=settings["cipher"],
                            key_size=settings["key_size"],
                        )
                    )
                except KeyError as e:
                    raise UserError(
                        f"Missing required key '{e.args[0]}' in encryption item '{i}'"
                    )
        return cls(encryptions=encryptions)

    @classmethod
    def from_system(cls, host: str) -> Self:
        from kisiac.filesystems import DeviceInfos

        encryptions = set()
        luks_devices = [
            device for device in DeviceInfos(host) if device.fstype == "crypto_LUKS"
        ]
        for luks_device in luks_devices:
            output = json.loads(
                run_cmd(
                    [
                        "cryptsetup",
                        "luksDump",
                        "--dump-json-metadata",
                        str(luks_device.device),
                    ],
                    sudo=True,
                ).stdout
            )
            if len(luks_device.children) > 1:
                raise UserError(
                    f"Unexpected number of children for LUKS device '{luks_device.device}', "
                    "expected at most 1. This means that your encryption setup is not yet "
                    "supported by kisiac."
                )
            name = (
                luks_device.children[0].device.name
                if len(luks_device.children) == 1
                else None
            )
            # TODO find a way to retrieve the key_size from the system
            encryptions.add(
                Encryption(
                    name=name,
                    device=luks_device.device,
                    hash=output["keyslots"]["0"]["af"]["hash"],
                    cipher=output["keyslots"]["0"]["area"]["encryption"],
                    key_size=None,
                )
            )
        return cls(encryptions=encryptions)

    def by_name(self) -> dict[str, Encryption]:
        return {
            encryption.name: encryption
            for encryption in self.encryptions
            if encryption.name is not None
        }

    def by_device(self) -> dict[Path, Encryption]:
        return {encryption.device: encryption for encryption in self.encryptions}

    def __iter__(self) -> Iterator[Encryption]:
        return iter(self.encryptions)



import sys
from kisiac.common import (
    HostAgnosticPath,
    UserError,
    cmd_to_str,
    confirm_action,
    log_action,
    provide_password,
    run_cmd,
)
from kisiac.encryption import EncryptionSetup
from kisiac.filesystems import DeviceInfos, update_filesystems
from kisiac.runtime_settings import GlobalSettings, UpdateHostSettings
from kisiac import users
from kisiac.config import Config
from kisiac.lvm import LVMSetup

import inquirer


default_system_software = [
    "openssh-server",
    "openssh-client",
    "lvm2",
    "e2fsprogs",
    "xfsprogs",
    "btrfs-progs",
]


def setup_config() -> None:
    if GlobalSettings.get_instance().non_interactive:
        content = sys.stdin.read()
    else:
        answers = inquirer.prompt(
            [
                inquirer.Text(
                    "secret_config",
                    message="Paste the secret configuration (YAML format), including the repo key",
                ),
            ]
        )
        assert answers is not None
        content = answers["secret_config"]
    HostAgnosticPath("/etc/kisiac.yaml", sudo=True).write_text(content)


def update_host(host: str) -> None:
    config = Config.get_instance()
    for file in config.files.get_files(user=None):
        log_action(host, "Updating system file", file.target_path)
        file.write(overwrite_existing=True, host=host, sudo=True)

    update_system_packages(host)

    update_encryptions(host)

    update_lvm(host)

    update_filesystems(host)

    users.setup_users(host=host)
    for user in config.users:
        for file in config.files.get_files(user.username):
            log_action(host, "Updating user file", file.target_path)
            # If the user already has the files, we leave him the new file as a
            # template next to the actual file, with the suffix '.updated'.
            user.fix_permissions(
                file.write(overwrite_existing=False, host=host, sudo=True), host=host
            )


def update_system_packages(host: str) -> None:
    run_cmd(["apt-get", "update"], sudo=True, host=host)
    if not UpdateHostSettings.get_instance().skip_system_upgrade:
        run_cmd(["apt-get", "upgrade"], sudo=True, host=host)
    run_cmd(
        ["apt-get", "install"]
        + list(set(Config.get_instance().system_software + default_system_software)),
        sudo=True,
        host=host,
    )


def update_encryptions(host: str) -> None:
    desired = Config.get_instance().encryption
    current = EncryptionSetup.from_system(host=host)

    desired_by_name = desired.by_name()
    current_by_device = current.by_device()
    current_by_name = current.by_name()

    for encryption in desired:
        if encryption.name is not None:
            curr_encryption = current_by_name.get(encryption.name)
            if curr_encryption is not None:
                if curr_encryption != encryption:
                    # TODO support such changes
                    raise UserError(
                        f"Encryption {encryption.name} has changed. "
                        "Modifying it via kisiac is not yet supported. "
                        f"Current: {curr_encryption}, Desired: {encryption}"
                    )

    dd_cmds = []
    format_cmds = []

    for encryption in desired:
        if encryption.device not in current_by_device:
            # overwrite the header with random data
            dd_cmds.append(
                [
                    "dd",
                    "if=/dev/urandom",
                    "bs=1M",
                    "count=8",
                    f"of={encryption.device!s}",
                ]
            )
            format_cmds.append(
                [
                    "cryptsetup",
                    "luksFormat",
                    "--cipher",
                    encryption.cipher,
                    "--key-size",
                    encryption.key_size,
                    "--hash",
                    encryption.hash,
                    "--key-file",
                    "-",
                    encryption.device,
                ]
            )

    cmd_msg = cmd_to_str(*(dd_cmds + format_cmds))

    password = None

    def get_password() -> str:
        return provide_password("Provide encryption password.")

    if format_cmds and confirm_action(
        f"The following cryptsetup commands will be executed:\n{cmd_msg}\n"
        "\nProceed? If answering no, consider making the changes manually or "
        "adjust the kisiac encryption configuration."
    ):
        error_msg = "Incomplete encryption update due to error (make sure to manually fix this!)"
        password = get_password()
        for cmd in dd_cmds:
            run_cmd(cmd, host=host, sudo=True, user_error_msg=error_msg)
        for cmd in format_cmds:
            run_cmd(cmd, host=host, sudo=True, user_error_msg=error_msg, input=password)

    encryptions_to_open = [
        encryption for encryption in desired if encryption.name not in current_by_name
    ]
    encryptions_to_reopen = [
        encryption
        for encryption in desired
        if encryption.name in current_by_name
        and current_by_name[encryption.name].device != encryption.device
    ]
    encryptions_to_close = [
        encryption for encryption in current if encryption.name not in desired_by_name
    ]

    if encryptions_to_open or encryptions_to_reopen or encryptions_to_close:
        if password is None:
            password = get_password()

        for encryption in encryptions_to_open:
            encryption.open(host, password)
        for encryption in encryptions_to_reopen:
            encryption.close(host)
            encryption.open(host, password)
        for encryption in encryptions_to_close:
            encryption.close(host)


def update_lvm(host: str) -> None:
    desired = Config.get_instance().lvm
    current = LVMSetup.from_system(host=host)
    device_infos = DeviceInfos(host)

    cmds = []

    cmds.extend(
        [
            "lvremove",
            "--yes",
            f"{vg.name}/{lv}",
        ]
        for vg in current.vgs.values()
        for lv in vg.lvs.values()
        if vg.name not in desired.vgs or lv.name not in desired.vgs[vg.name].lvs
    )
    cmds.extend(
        ["vgremove", "--yes", vg] for vg in current.vgs.keys() - desired.vgs.keys()
    )
    pvremove = [pv.device for pv in current.pvs - desired.pvs]
    if pvremove:
        cmds.append(["pvremove", "--yes", *pvremove])

    pvcreate = [pv.device for pv in desired.pvs - current.pvs]
    if pvcreate:
        cmds.append(["pvcreate", "--yes", *pvcreate])
    cmds.extend(
        ["vgcreate", vg.name] + [pv.device for pv in vg.pvs]
        for vg_name, vg in desired.vgs.items()
        if vg_name not in current.vgs
    )
    cmds.extend(
        [
            "lvcreate",
            "-n",
            lv.name,
            *lv.size_arg(),
            vg.name,
            "--type",
            lv.layout,
        ]
        for vg in desired.vgs.values()
        for lv in vg.lvs.values()
        if vg.name not in current.vgs or lv.name not in current.vgs[vg.name].lvs
    )

    # update existing VGs and LVs
    for vg_desired in desired.vgs.values():
        vg_current = current.vgs.get(vg_desired.name)
        if vg_current is None:
            continue

        # update pvs in vg
        pvs_to_add = vg_desired.pvs - vg_current.pvs
        if pvs_to_add:
            cmds.append(
                ["vgextend", vg_desired.name] + [pv.device for pv in pvs_to_add]
            )
        pvs_to_remove = vg_current.pvs - vg_desired.pvs
        if pvs_to_remove:
            cmds.append(
                ["vgreduce", "--yes", vg_desired.name]
                + [pv.device for pv in pvs_to_remove]
            )

        # update lvs in vg
        for lv_desired in vg_desired.lvs.values():
            lv_current = vg_current.lvs.get(lv_desired.name)
            if lv_current is None:
                continue
            if lv_current.layout != lv_desired.layout:
                raise UserError(
                    f"Cannot change layout of existing LV {lv_desired.name} "
                    f"from {lv_current.layout} to {lv_desired.layout}. "
                    "Perform this action manually and re-run the update."
                )

            if not lv_current.is_same_size(lv_desired):
                if lv_desired.fills_vg():
                    log_action(host, f"Ensuring that LV {lv_desired.name} fills VG.")
                else:
                    log_action(
                        host,
                        f"Resizing LV {lv_desired.name} from {lv_current.size} to "
                        f"{lv_desired.size}",
                    )

                device_info = device_infos.get_info_for_device(
                    vg_desired.get_lv_device(lv_desired.name)
                )
                resize_fs = ["--resizefs"] if device_info.fstype is not None else []

                cmds.append(
                    [
                        "lvresize",
                        *resize_fs,
                        *lv_desired.size_arg(),
                        f"{vg_desired.name}/{lv_desired.name}",
                    ]
                )
    cmd_msg = cmd_to_str(*cmds)

    if confirm_action(
        f"The following LVM commands will be executed:\n{cmd_msg}\n"
        "\nProceed? If answering no, consider making the changes manually or "
        "adjust the kisiac LVM configuration."
    ):
        for cmd in cmds:
            run_cmd(
                cmd,
                host=host,
                sudo=True,
                user_error_msg="Incomplete LVM update due to error (make sure to manually fix this!)",
            )



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
import platform
import re
from typing import Any, Iterable, Self, Sequence
import base64

import jinja2
from kisiac.encryption import EncryptionSetup
import yaml
import git
from pyfstab.entry import Entry as FstabEntry
import yte

from kisiac.common import (
    HostAgnosticPath,
    Singleton,
    cache,
    UserError,
    check_type,
    handle_key_error,
)
from kisiac.lvm import LVMSetup


config_file_path = Path("/etc/kisiac.yaml")


required_marker = object()


@dataclass
class Package:
    name: str
    cmd_spec: str | None
    desc: str
    with_pkgs: list[str]
    post_install: str | None

    @property
    def cmd(self) -> str:
        return self.cmd_spec or self.name

    @property
    def install_cmd(self) -> str:
        supplement = " ".join(f"--with {pkg}" for pkg in self.with_pkgs)

        if self.post_install:
            supplement += " && ".join(self.post_install.splitlines())

        return f"pixi global install {self.name} {supplement}"


class FileType(Enum):
    system = "system"
    user = "user"


@dataclass(frozen=True, order=True)
class Filesystem:
    device: Path | None
    label: str | None
    uuid: str | None
    fstype: str
    mountpoint: Path | None
    options: str | None
    dump: int
    fsck: int

    def __post_init__(self) -> None:
        if (
            sum(1 for item in (self.device, self.label, self.uuid) if item is not None)
            != 1
        ):
            raise UserError("Filesystem may only have one of device, label or uuid.")

    @classmethod
    def from_fstab_entry(cls, entry: FstabEntry) -> Self:
        assert entry.device is not None
        assert entry.type is not None
        if entry.device.startswith("LABEL="):
            device = None
            label = entry.device[len("LABEL=") :]
            uuid = None
        elif entry.device.startswith("UUID="):
            device = None
            label = None
            uuid = entry.device[len("UUID=") :]
        else:
            device = Path(entry.device)
            label = None
            uuid = None
        return cls(
            device=device,
            label=label,
            uuid=uuid,
            fstype=entry.type,
            mountpoint=Path(entry.dir) if entry.dir is not None else None,
            options=entry.options,
            dump=entry.dump or 0,
            fsck=entry.fsck or 0,
        )

    def to_fstab_entry(self) -> FstabEntry:
        return FstabEntry(
            _device=str(self.device or self.label or self.uuid),
            _dir=str(self.mountpoint) if self.mountpoint is not None else None,
            _type=self.fstype,
            _options=self.options,
            _dump=self.dump,
            _fsck=self.fsck,
        )


class UserSet(Enum):
    nobody = "nobody"
    owner = "owner"
    group = "group"
    others = "others"


@dataclass
class Permissions:
    owner: str | None
    group: str | None
    read: UserSet | None
    write: UserSet | None
    execute: UserSet | None
    setgid: bool
    setuid: bool
    sticky: bool


@dataclass
class File:
    target_path: Path
    content: str

    def write(self, overwrite_existing: bool, host: str, sudo: bool) -> Sequence[Path]:
        target_path = HostAgnosticPath(self.target_path, host=host, sudo=sudo)
        if target_path.exists():
            if target_path.read_text() == self.content:
                return []
            if not overwrite_existing:
                target_path = target_path.with_suffix(".updated")
        created = []
        for ancestor in target_path.parents[::-1][1:]:
            if not ancestor.exists():
                ancestor.mkdir()
                created.append(ancestor.path)
        target_path.write_text(self.content)
        created.append(target_path.path)
        return created


class Files:
    def __init__(self, config: "Config") -> None:
        cache_address = base64.b64encode(config.repo.encode()).decode()
        self.repo_cache = cache / cache_address
        self.infrastructure = config.infrastructure
        self.vars = config.vars
        self.user_vars = config.user_vars
        if not self.repo_cache.exists():
            self.repo_cache.parent.mkdir(parents=True, exist_ok=True)
            self.repo = git.Repo.clone_from(config.repo, self.repo_cache)
        else:
            self.repo = git.Repo(self.repo_cache)
            if self.repo.remotes:
                # update to latest commit
                self.repo.remotes.origin.pull()

    def infrastructure_stack(self) -> Iterable[Path]:
        base = self.repo_cache / "infrastructure"
        all_path = base / "all"
        if all_path.exists():
            yield base / "all"
        if self.infrastructure is not None:
            infra_path = base / self.infrastructure
            if infra_path.exists():
                yield infra_path

    def host_stack(self) -> Iterable[Path]:
        hostname = platform.node()
        for infra in self.infrastructure_stack():
            base = infra / "hosts"
            if base.exists():
                for entry in base.iterdir():
                    if not entry.is_dir():
                        raise UserError(f"{base} may only contain directories")
                    # yield if all or entry matches hostname
                    regex = str(entry).replace("*", r".+")
                    if entry.name == "all" or re.match(regex, hostname):
                        yield entry

    def get_config(self) -> dict[str, Any]:
        config = {}
        for base in self.host_stack():
            config_path = base / "kisiac.yaml"
            if config_path.exists():
                with open(config_path, "r") as f:
                    config.update(load_config(f))
        return config

    def get_files(self, user: str | None) -> Iterable[File]:
        if user is not None:
            file_type = "user_files"
            vars = dict(self.vars) | self.user_vars(user)

            # yield built-in user files
            templates = jinja2.Environment(
                loader=jinja2.PackageLoader("kisiac", "files"),
                autoescape=jinja2.select_autoescape(),
            )
            content = templates.get_template("kisiac.sh.j2").render(
                packages=Config.get_instance().user_software,
                infrastructure_name=Config.get_instance().infrastructure_name,
                infrastructure_name_len=len(Config.get_instance().infrastructure_name),
                messages=Config.get_instance().messages,
            )
            yield File(target_path=Path("/etc/profile.d/kisiac.sh"), content=content)
        else:
            file_type = "system_files"
            vars = self.vars

        for host in self.host_stack():
            collection = host / file_type
            templates = jinja2.Environment(
                loader=jinja2.FileSystemLoader(host),
                autoescape=jinja2.select_autoescape(),
            )
            for base, _, files in (collection).walk():
                for f in files:
                    if f.endswith(".j2"):
                        content = templates.get_template(str(base / f)).render(**vars)
                    elif f.endswith(".yaml"):
                        with open(base / f, "r") as fileobj:
                            content = yte.process_yaml(
                                fileobj, variables=vars, require_use_yte=True
                            )
                        assert content is not None
                        content = yaml.dump(content, indent=2)
                    else:
                        with open(base / f, "r") as content:
                            content = content.read()
                    yield File((base / f).relative_to(collection), content)


@dataclass
class User:
    username: str
    primary_group: str
    secondary_groups: list[str]
    ssh_pub_key: str
    vars: dict[str, Any]

    def fix_permissions(self, paths: Iterable[Path], host: str) -> None:
        for path in paths:
            path = HostAgnosticPath(path, host=host, sudo=True)
            # ensure that only user may read/write the paths
            if path.is_dir():
                path.chmod("u=rwx", "g-rwx", "o-rwx")
            else:
                path.chmod("u=rw", "g-rwx", "o-rwx")
            path.chown(self.username, self.primary_group)


class Config(Singleton):
    def __init__(self) -> None:
        # Config is bootstrapped via an env variable that contains YAML or the file config_file_path.
        # It at least has to contain the repo: ... key that points to a
        # git repo containing the actual config.
        # However, it may also contain e.g. user definitions or secret (environment)
        # variables.

        self._config: dict[str, Any] = {}

        config_set = False

        try:
            with open(config_file_path, "r") as f:
                self._config.update(load_config(f))
            config_set = True
        except (FileNotFoundError, IOError):
            # ignore missing file or read errors, we fall back to env var
            pass
        except Exception as e:
            # raise other errors
            raise UserError(f"Error reading config file {config_file_path}: {e}") from e

        if not config_set:
            raise UserError(
                "No config file found at "
                f"{config_file_path}. Run 'kisiac setup-config' to set up the "
                "configuration."
            )

        self._files: Files | None = None

        self._config.update(self.files.get_config())

    def as_str(self) -> str:
        return yaml.dump(self._config)

    def get(self, key: str, default: Any | None = required_marker) -> Any:
        value = self._config.get(key, default)

        if value is required_marker:
            raise UserError(f"Config lacks key {key}.")

        return value

    @property
    @handle_key_error("Invalid user configuration")
    def users(self) -> Iterable[User]:
        users = self.get("users")
        check_type("users key", users, dict)

        for username, settings in users.items():
            check_type(f"user {username}", settings, dict)
            primary_group = settings["groups"]["primary"]
            secondary_groups = settings["groups"].get("secondary", [])
            check_type(f"user {username} groups", secondary_groups, list)
            yield User(
                username,
                ssh_pub_key=settings["ssh_pub_key"],
                vars=settings.get("vars", {}),
                primary_group=primary_group,
                secondary_groups=list(map(str, secondary_groups)),
            )

    @property
    def vars(self) -> dict[str, Any]:
        vars = self.get("vars", default={})
        check_type("vars key", vars, dict)
        return vars

    def user_vars(self, user: str) -> dict[str, Any]:
        return self.get("users", default={})[user].get("vars", {})

    @property
    def infrastructure(self) -> str:
        infrastructure = self.get("infrastructure", default=None)
        check_type("infrastructure key", infrastructure, (str, type(None)))
        return infrastructure

    @property
    def repo(self) -> str:
        repo_url = self.get("repo")
        check_type("repo key", repo_url, str)
        return repo_url

    @property
    def files(self) -> Files:
        if self._files is None:
            self._files = Files(self)

        return self._files

    @property
    def user_software(self) -> Iterable[Package]:
        user_software = self.get("user_software")
        check_type("user_software key", user_software, list)
        for entry in user_software:
            check_type("user_software entry", entry, dict)
            try:
                yield Package(
                    name=entry["pkg"],
                    cmd_spec=entry.get("cmd"),
                    desc=entry["desc"],
                    with_pkgs=entry.get("with", []),
                    post_install=entry.get("post_install"),
                )
            except KeyError as e:
                raise UserError(f"Missing {e} in user_software definition.")

    @property
    def system_software(self) -> list[str]:
        system_software = self.get("system_software", default=[])
        check_type("system_software key", system_software, list)
        return system_software

    @property
    def messages(self) -> Sequence[str]:
        messages = self.get("messages", default=[])
        check_type("message key", messages, list)
        return messages

    @property
    def infrastructure_name(self) -> str:
        infrastructure_name = self.get("infrastructure_name")
        check_type("infrastructure_name key", infrastructure_name, str)
        return infrastructure_name

    @property
    def encryption(self) -> EncryptionSetup:
        encryption = self.get("encryption", default={})
        return EncryptionSetup.from_config(encryption)

    @property
    def lvm(self) -> LVMSetup:
        lvm = self.get("lvm", default={})
        return LVMSetup.from_config(lvm)

    @property
    def filesystems(self) -> list[Filesystem]:
        filesystems = self.get("filesystems", default={})
        check_type("filesystems key", filesystems, list)

        entries = []
        for settings in filesystems:
            check_type("filesystem item", settings, dict)
            device = settings.get("device")
            entries.append(
                Filesystem(
                    device=Path(device) if device is not None else None,
                    label=settings.get("label"),
                    uuid=settings.get("uuid"),
                    fstype=settings["type"],
                    mountpoint=settings["mount"],
                    options=settings.get("options", ""),
                    dump=settings.get("dump", 0),
                    fsck=settings.get("pass", 2),
                )
            )
        if any(filesystem.fstype == "swap" for filesystem in entries) and any(
            self.encryption
        ):
            raise UserError(
                "Swap partition set up, but physicial volume with encryption specified. "
                "This is a risk since encrypted data can end up in the swap. "
                "It is possible to encrypt swap as well, but this is currently not supported "
                "by kisiac. Remove the swap partition or disable encryption."
            )
        return entries

    @property
    def permissions(self) -> dict[Path, Permissions]:
        check_type("permissions key", self._config.get("permissions", {}), dict)
        permissions = {}
        for path_str, settings in self._config.get("permissions", {}).items():
            check_type(f"permissions for {path_str}", settings, dict)
            path = Path(path_str)
            permissions[path] = Permissions(
                owner=settings.get("owner"),
                group=settings.get("group"),
                read=UserSet(settings["read"]) if "read" in settings else None,
                write=UserSet(settings["write"]) if "write" in settings else None,
                execute=UserSet(settings["execute"]) if "execute" in settings else None,
                setgid=settings.get("setgid", False),
                setuid=settings.get("setuid", False),
                sticky=settings.get("sticky", False),
            )
        return permissions


def load_config(config) -> dict[Any, Any]:
    config = yte.process_yaml(config, require_use_yte=True)

    if not isinstance(config, dict):
        raise ValueError("Config has to be a mapping")
    return config



from dataclasses import dataclass, field
import json
from pathlib import Path
from typing import Any, Self

from humanfriendly import parse_size

from kisiac.common import check_type, exists_cmd, run_cmd


CRYPT_PREFIX = "crypt_"


@dataclass(frozen=True)
class PV:
    device: str


@dataclass(frozen=True)
class LV:
    name: str
    layout: str
    size: int | None

    def is_same_size(self, other: Self) -> bool:
        if self.fills_vg() or other.fills_vg():
            # if both just fill the VG, their sizes are considered unchanged
            return self.fills_vg() and other.fills_vg()

        assert self.size is not None and other.size is not None

        def simplify(size: int) -> int:
            # tens of MB should be precise enough
            return size // 10**7

        return simplify(self.size) == simplify(other.size)

    def fills_vg(self) -> bool:
        return self.size is not None

    def size_arg(self) -> list[str]:
        if self.size is None:
            return ["--extents", "+100%FREE"]
        else:
            return ["--size", f"{self.size}B"]


@dataclass(frozen=True)
class VG:
    name: str
    pvs: set[PV] = field(default_factory=set)
    lvs: dict[str, LV] = field(default_factory=dict)

    def get_lv_device(self, lv_name: str) -> Path:
        return Path("/dev") / self.name / lv_name


@dataclass
class LVMSetup:
    pvs: set[PV] = field(default_factory=set)
    vgs: dict[str, VG] = field(default_factory=dict)

    def is_empty(self) -> bool:
        return not self.pvs and not self.vgs

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> Self:
        check_type("lvm key", config, dict)
        entities = cls()
        for pv in config.get("pvs", []):
            check_type("lvm pv entry", pv, str)
            entities.pvs.add(
                PV(
                    device=pv,
                )
            )
        for name, settings in config.get("vgs", {}).items():
            check_type(f"lvm vg {name} entry", settings, dict)

            lvs = settings.get("lvs", {})
            check_type(f"lvm vg {name} lvs entry", lvs, dict)

            lvs_entities = {}
            for lv_name, lv_settings in lvs.items():
                check_type(f"lvm vg {name} lv {lv_name} entry", lv_settings, dict)

                size = lv_settings["size"]
                if size == "rest":
                    size = None
                else:
                    size = parse_size(size, binary=True)

                lvs_entities[lv_name] = LV(
                    name=lv_name,
                    layout=lv_settings["layout"],
                    size=size,
                )

            entities.vgs[name] = VG(
                name=name,
                pvs={PV(device=pv) for pv in settings.get("pvs", [])},
                lvs=lvs_entities,
            )
        return entities

    @classmethod
    def from_system(cls, host: str) -> Self:
        # check if lvm2 is installed, return empty LVM entities otherwise
        if not exists_cmd("pvcreate", host=host, sudo=True):
            return cls()

        entities: Self = cls()

        lv_data = json.loads(
            run_cmd(
                [
                    "lvs",
                    "--units",
                    "b",
                    "--options",
                    "lv_name,vg_name,lv_layout,lv_size",
                    "--reportformat",
                    "json",
                ],
                host=host,
                sudo=True,
            ).stdout
        )["report"][0]["lv"]

        vg_data = json.loads(
            run_cmd(
                ["vgs", "--options", "vg_name", "--reportformat", "json"],
                host=host,
                sudo=True,
            ).stdout
        )["report"][0]["vg"]

        pv_data = json.loads(
            run_cmd(
                ["pvs", "--options", "pv_name,vg_name", "--reportformat", "json"],
                host=host,
                sudo=True,
            ).stdout
        )["report"][0]["pv"]

        for entry in vg_data:
            entities.vgs[entry["vg_name"]] = VG(name=entry["vg_name"])
        for entry in pv_data:
            pv_device = entry["pv_name"]

            pv_obj = PV(device=pv_device)
            entities.pvs.add(pv_obj)
            entities.vgs[entry["vg_name"]].pvs.add(pv_obj)

        for entry in lv_data:
            vg = entities.vgs[entry["vg_name"]]
            vg.lvs[entry["lv_name"]] = LV(
                name=entry["lv_name"],
                layout=entry["lv_layout"],
                size=parse_size(entry["lv_size"], binary=True),
            )
        return entities
